// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  ifgif.spi - GIF import plug-in
//  (C) 2004 mij4x
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  gif_img.cpp
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "gif_img.h"
#include <string.h>


// ----------------------------------------------------------------------------
//  vg^Cv錾
// ----------------------------------------------------------------------------

int check_header(const gif_header *hdr);
int check_image_descriptor(const image_descriptor *img_desc);
const uint8 *skip_to_image(const uint8 *img, uint size);


// ----------------------------------------------------------------------------
//  gif 摜t@C
// ----------------------------------------------------------------------------

void copy_row_bin(uint8 *dst, const uint8 *src, int count)
{
  uint8  *dst_p = dst;
  const uint8  *src_p = src;
  for (int i = 0; i < count - 7; i += 8, src_p += 8)
    *dst_p++ = (src_p[0] << 7) | 
               (src_p[1] << 6) | 
               (src_p[2] << 5) | 
               (src_p[3] << 4) | 
               (src_p[4] << 3) | 
               (src_p[5] << 2) | 
               (src_p[6] << 1) | 
               (src_p[7]     );

  if (count & 0x07)
  {
    for (int i = 0; i < (count & 0x07); i++)
      *dst_p = (*dst_p << 1) | *src_p++;
    *dst_p <<= 8 - (count & 0x07);
  }
}

void copy_row_16(uint8 *dst, const uint8 *src, int count)
{
  for (int i = 0; i < count - 1; i += 2)
    dst[i >> 1] = (src[i] << 4) | src[i + 1];

  if (count & 1)
    dst[count >> 1] = src[count - 1] << 4;
}

void copy_row_256(uint8 *dst, const uint8 *src, int count)
{
  memcpy(dst, src, count);
}


// ---
//  LZW fR[h̉摜 DIB ƂăRs[
// ---

// V[PV
void gif_image::copy_sequencial(uint8 *dst, const uint8 *src, copy_func cp_row) const
{
  int  gif_row_bytes = get_width();
  int  dib_row_bytes = ((get_width() * get_color_depth() + 31) >> 3) & 0xfffffffcUL;
  int  height = get_height();

  for (int i = 0; i < height; i++)
    cp_row(dst + dib_row_bytes * (height - 1 - i),
           src + gif_row_bytes * i,
           gif_row_bytes);
}

// C^[X
void gif_image::copy_interlace(uint8 *dst, const uint8 *src, copy_func cp_row) const
{
  int  gif_row_bytes = get_width();
  int  dib_row_bytes = ((get_width() * get_color_depth() + 31) >> 3) & 0xfffffffcUL;
  int  height = get_height();

  int  j = 0;
  for (int i = 0; i < height; i += 8)
    cp_row(dst + dib_row_bytes * (height - 1 - i),
           src + gif_row_bytes * j++,
           gif_row_bytes);

  for (int i = 4; i < height; i += 8)
    cp_row(dst + dib_row_bytes * (height - 1 - i),
           src + gif_row_bytes * j++,
           gif_row_bytes);

  for (int i = 2; i < height; i += 4)
    cp_row(dst + dib_row_bytes * (height - 1 - i),
           src + gif_row_bytes * j++,
           gif_row_bytes);

  for (int i = 1; i < height; i += 2)
    cp_row(dst + dib_row_bytes * (height - 1 - i),
           src + gif_row_bytes * j++,
           gif_row_bytes);
}


// ---
//  gif 摜J
// ---

int gif_image::open(const uint8 *file_img, uint file_size)
{
  close();
  set_error_code(SPI_ERROR_BROKEN_DATA);

  // I[|C^߂
  const uint8  *file_img_end = file_img + file_size;

  // gif wb_Zbg
  m_hdr = (const gif_header *)file_img;
  if (file_size < sizeof(gif_header) || check_header(m_hdr))
    return 1;

  const uint8  *p = file_img + sizeof(gif_header);

  // O[oJ[e[uȂC̃|C^L^
  if (GET_GLOBAL_COLOR_TABLE_FLAG(m_hdr->packed_fields))
  {
    if (file_img_end < (p = set_color_table(p, GET_GCT_SIZE(m_hdr->packed_fields) + 1)))
      return 1;
  }

  // ŏ image descriptor ܂œǂݔ΂
  if (NULL == (p = skip_to_image(p, file_img_end - p)))
    return 1;

  m_img_desc = (const image_descriptor *)p;
  if (check_image_descriptor(m_img_desc))
    return 1;

  p += sizeof(image_descriptor);

#if 0
  if (m_hdr->width < m_img_desc->width ||
      m_hdr->height < m_img_desc->height)
    return 1;
#endif

  // [JJ[e[uȂC̃|C^L^ (炪D)
  if (ID_GET_LOCAL_COLOR_TABLE_FLAG(m_img_desc->packed_fields))
  {
    if (file_img_end < (p = set_color_table(p, ID_GET_LCT_SIZE(m_img_desc->packed_fields) + 1)))
      return 1;
  }

  // J[e[uȂH
  if (NULL == m_clr_tbl)
    return 1;

  // LZW kꂽC[W̃|C^L^
  m_lzw_img = p;
  m_lzw_img_size = file_img_end - p;

  set_error_code(SPI_OK);

  return 0;
}


// ---
//  摜擾
// ---

int gif_image::get_image(uint8 *out_buf, SPI_PROGRESS_FUNC prog_func, long pf_data)
{
  size_t  buf_size = get_width() * get_height();
  uint8  *buf;
  try {
    buf = new uint8 [buf_size + GIF_LZW_MAX_TOKEN_COUNT + 10];
  } catch (...) {
    set_error_code(SPI_ERROR_MEMORY_ALLOCATION);
    return 1;
  }

  m_prog_func = prog_func;
  m_pf_data = pf_data;

#ifdef ENABLE_PROGRESS_CALLBACK
  if (NULL != m_prog_func) m_prog_func(0, m_lzw_img_size, m_pf_data);
#endif

  int ret = decode_gif_lzw(m_lzw_img, m_lzw_img_size,
                           buf, buf_size);

#ifdef ENABLE_PROGRESS_CALLBACK
  if (SPI_OK == ret && NULL != m_prog_func)
    m_prog_func(m_lzw_img_size, m_lzw_img_size, m_pf_data);
#endif

  if (!ret)
  {
    copy_func  cp_row;
    switch (get_color_depth())
    {
    case 1: cp_row = copy_row_bin; break;
    case 4: cp_row = copy_row_16; break;
    case 8: cp_row = copy_row_256; break;
    default: cp_row = NULL;
    }

    if (NULL != cp_row)
    {
      if (!ID_GET_INTERLACE_FLAG(m_img_desc->packed_fields))
        copy_sequencial(out_buf, buf, cp_row);
      else
        copy_interlace(out_buf, buf, cp_row);
    }
    else
      ret = 1;
  }

  delete [] buf;

  set_error_code(ret);

  return ret;
}


// ----------------------------------------------------------------------------
//  Xg[
// ----------------------------------------------------------------------------

class memory_stream
{
  const uint8  *m_ptr;
  const uint8  *m_ptr_end;

public:
  memory_stream(const uint8 *mem, size_t size) { set_memory(mem, size); }
  ~memory_stream() { }

  void set_memory(const uint8 *mem, size_t size)
  {
    m_ptr = mem;
    m_ptr_end = mem + size;
  }

  const uint8 *get_pointer(void) const { return m_ptr; }

  uint8 byte(void)
  {
    if (m_ptr >= m_ptr_end)
      throw;
    return *m_ptr++;
  }
};


// ----------------------------------------------------------------------------
//  GIF wb_ُ̈𒲂ׂ
// ----------------------------------------------------------------------------

int check_header(const gif_header *hdr)
{
  static const char  *GIF_SIGNATURE = "GIF8";

  if (NULL == hdr ||
      memcmp(GIF_SIGNATURE, hdr->sig_ver, strlen(GIF_SIGNATURE)) ||  // "GIF8"
      ('7' != hdr->sig_ver[4] && '9' != hdr->sig_ver[4]) ||
      'a' != hdr->sig_ver[5] ||
      0 == hdr->width || 0 == hdr->height)
    return 1;

#if 0
  uint  num_colors = 1 << (GET_COLOR_RESOLUTION(hdr->packed_fields) + 1);
  uint  gct_num_colors = 1 << (GET_GCT_SIZE(hdr->packed_fields) + 1);

  if (num_colors < gct_num_colors)
    return 1;
#endif

  return 0;
}


// ----------------------------------------------------------------------------
//  C[WfBXNv^ُ̈𒲂ׂ
// ----------------------------------------------------------------------------

int check_image_descriptor(const image_descriptor *img_desc)
{
  if (img_desc->width <= img_desc->x_pos ||
      img_desc->height <= img_desc->y_pos)
    return 1;
  else
    return 0;
}


// ----------------------------------------------------------------------------
//  C[WfBXNv^܂œǂݔ΂
// ----------------------------------------------------------------------------

#define SKIP_BYTES(m, c) \
  do { \
    for (int i = 0; i < (c); i++) \
      (m).byte(); \
  } while (0)

#define SKIP_SUB_BLOCKS(m) \
  do { \
    for (int l = (m).byte(); l > 0; l = (m).byte()) \
      while (l--) \
        (m).byte(); \
  } while (0)

const uint8 *skip_to_image(const uint8 *img, size_t size)
{
  memory_stream  strm(img, size);

  try
  {
    while (1)
    {
      switch (strm.byte())
      {
      case 0x2c:  // image descriptor
        return strm.get_pointer() - 1;

      case 0x3b:  // trailer
        return NULL;

      case 0x21:  // extension
        switch (strm.byte())
        {
        case 0xf9:  // graphic control extension
          if (4 != strm.byte()) return NULL;
          SKIP_BYTES(strm, 4);
          SKIP_SUB_BLOCKS(strm);
          break;

        case 0x01:  // plain text extension
          if (12 != strm.byte()) return NULL;
          SKIP_BYTES(strm, 12);
          SKIP_SUB_BLOCKS(strm);
          break;

        case 0xff:  // application extension
          if (11 != strm.byte()) return NULL;
          SKIP_BYTES(strm, 11);
          SKIP_SUB_BLOCKS(strm);
          break;

        case 0xfe:  // comment extension
          SKIP_SUB_BLOCKS(strm);
          break;
        }
        break;
      }
    }
  }
  catch (...)
  {
    return NULL;
  }
}


// ----------------------------------------------------------------------------
//  gif LZW ̓WJs
// ----------------------------------------------------------------------------

struct lzw_token
{
  const uint8  *begin;
  const uint8  *end;
};

int gif_image::decode_gif_lzw(const uint8 *src_img, size_t src_size,
                              uint8 *dst_buf, size_t buf_size)
{
  static const uint BIT_MASK[] =
  {
    0,
    0x0001, 0x0003, 0x0007, 0x000f, 0x001f, 0x003f, 0x007f, 0x00ff,
    0x01ff, 0x03ff, 0x07ff, 0x0fff, 0x1fff, 0x3fff, 0x7fff, 0xffff,
  };

  lzw_token  token[gif_image::GIF_LZW_MAX_TOKEN_COUNT + 1];

  uint8        *out_p = dst_buf;
  const uint8  *out_end = dst_buf + buf_size;

  const uint8  *src_p = src_img + 1;
  const uint8  *src_end_minus_one = src_img + src_size - 1;

  uint  init_bit_count = *src_img + 1;
  uint  num_literals = 1 << (init_bit_count - 1);
  uint  token_code_begin = num_literals + 2;

  uint  bitbuf = 0;
  uint  bitbuf_bits = 0;
  uint  block_bytes = 0;

  const uint8  *copy_src, *copy_end;
  uint  cur_code_bits, cur_bit_mask, v, i;

#define FETCH_BYTE \
  do { \
    if (0 == block_bytes) { \
      if (src_end_minus_one - src_p <= (int)(block_bytes = *src_p)) \
        return SPI_ERROR_BROKEN_DATA; \
      src_p++; \
    } \
    bitbuf |= *src_p++ << bitbuf_bits; \
    bitbuf_bits += 8; \
    block_bytes--; \
  } while (0)

#define GET_BITS \
  do { \
    v = bitbuf & cur_bit_mask; \
    bitbuf_bits -= cur_code_bits; \
    bitbuf >>= cur_code_bits; \
  } while (0)

#define COPY_TOKEN \
  do { \
    if (i <= v) return SPI_ERROR_BROKEN_DATA; \
    copy_src = token[v].begin; \
    copy_end = token[v].end; \
    *out_p++ = *copy_src++; \
    do { *out_p++ = *copy_src++; } while (copy_src < copy_end); \
    token[i].end = out_p + 1; \
  } while (0)

#define COPY_LITERAL \
  do { \
    *out_p++ = v; \
    token[i].end = out_p + 1; \
  } while (0)

  do {
    cur_code_bits = init_bit_count;
    cur_bit_mask = BIT_MASK[cur_code_bits];
    i = token_code_begin;
    if (init_bit_count <= 8) {
      while (out_p < out_end) {  // cur_code_bits  8 bit ܂ł̃[v
        if (bitbuf_bits < cur_code_bits) { FETCH_BYTE; }
        GET_BITS;
        token[i].begin = out_p;

        if (token_code_begin <= v) { COPY_TOKEN; }
        else if (v < num_literals) { COPY_LITERAL; }
        else if (v == num_literals)
        {
          cur_code_bits = init_bit_count;
          cur_bit_mask = BIT_MASK[cur_code_bits];
          i = token_code_begin - 1;
        }
        else return SPI_OK;

        if (i++ >> cur_code_bits) {
          cur_bit_mask = BIT_MASK[++cur_code_bits];
          if (cur_code_bits >= 9)
            break;
        }
      }
    }

    while (out_p < out_end) {  // 9 bit ȍ~̃[v
      FETCH_BYTE;
      if (bitbuf_bits < cur_code_bits) { FETCH_BYTE; }
      GET_BITS;

      token[i].begin = out_p;
      if (token_code_begin <= v) { COPY_TOKEN; }
      else if (v < num_literals) { COPY_LITERAL; }
      else if (v == num_literals) break;
      else return SPI_OK;

      if (i++ >> cur_code_bits) {
        if (cur_code_bits < gif_image::GIF_LZW_MAX_TOKEN_BITS)
          cur_bit_mask = BIT_MASK[++cur_code_bits];
        else
          i--;
      }
    }

#ifdef ENABLE_PROGRESS_CALLBACK
    if (NULL != m_prog_func && 0 != m_prog_func(src_p - src_img, src_size, m_pf_data))
      return SPI_ERROR_CANCEL;
#endif // ENABLE_PROGRESS_CALLBACK

  } while (out_p < out_end);

  return SPI_OK;
}


// ----------------------------------------------------------------------------
//  gif t@Cׂ
// ----------------------------------------------------------------------------

int check_image(const uint8 *data, uint data_size)
{
  if (data_size <= sizeof(gif_header))
    return 1;
  return check_header((const gif_header *)data);
}
