// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  gif_img.h
// ----------------------------------------------------------------------------

#ifndef gif_img_h__
#define gif_img_h__

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "conf.h"
#include "spi_obj.h"
#include "spi.h"


// ----------------------------------------------------------------------------
//  }N
// ----------------------------------------------------------------------------

#define GET_GLOBAL_COLOR_TABLE_FLAG(x)  (0x80 & (x))
#define GET_COLOR_RESOLUTION(x)  ((0x70 & (x)) >> 4)
#define GET_SORT_FLAG(x)  (0x08 & (x))
#define GET_GCT_SIZE(x)  (0x07 & (x))

#define ID_GET_LOCAL_COLOR_TABLE_FLAG(x)  (0x80 & (x))
#define ID_GET_INTERLACE_FLAG(x)  (0x40 & (x))
#define ID_GET_SORT_FLAG(x)  (0x20 & (x))
#define ID_GET_LCT_SIZE(x)  (0x07 & (x))


// ----------------------------------------------------------------------------
//  ^`
// ----------------------------------------------------------------------------

#pragma pack(1)

struct gif_header
{
  uint8   sig_ver[6];
  uint16  width;
  uint16  height;
  uint8   packed_fields;
  uint8   bg_color_index;
  uint8   pixel_aspect_ratio;
};

struct image_descriptor
{
  uint8   descriptor_id;
  uint16  x_pos;
  uint16  y_pos;
  uint16  width;
  uint16  height;
  uint8   packed_fields;
};

#pragma pack()


// ----------------------------------------------------------------------------
//  gif 摜t@C
// ----------------------------------------------------------------------------

class gif_image : public spi_object
{
  typedef void (*copy_func)(uint8 *, const uint8 *, int);

  const struct gif_header        *m_hdr;
  const struct image_descriptor  *m_img_desc;

  const uint8  *m_clr_tbl;
  uint         m_color_depth;

  const uint8  *m_lzw_img;
  uint         m_lzw_img_size;

  SPI_PROGRESS_FUNC  m_prog_func;
  long               m_pf_data;

  const uint8 *set_color_table(const uint8 *clr_tbl, uint color_depth)
  {
    m_color_depth = color_depth;
    m_clr_tbl = clr_tbl;
    return clr_tbl + (1 << m_color_depth) * 3;
  }

  void copy_sequencial(uint8 *dst, const uint8 *src, copy_func cp_row) const;
  void copy_interlace(uint8 *dst, const uint8 *src, copy_func cp_row) const;

  int decode_gif_lzw(const uint8 *src_img, size_t src_size,
                     uint8 *dst_buf, size_t buf_size);

public:
  enum
  {
    GIF_LZW_MAX_TOKEN_BITS  = 12,
    GIF_LZW_MAX_TOKEN_COUNT = 1 << GIF_LZW_MAX_TOKEN_BITS
  };

  gif_image(void) { close(); }
  ~gif_image() { }

  int open(const uint8 *file_img, uint file_size);

  void close(void)
  {
    m_hdr = NULL;
    m_img_desc = NULL;
    m_clr_tbl = NULL;
    m_lzw_img = NULL;
    m_color_depth = m_lzw_img_size = 0;
  }

  uint get_width(void) const { return m_img_desc->width; }
  uint get_height(void) const { return m_img_desc->height; }

  uint get_color_depth(void) const
  {
    switch (m_color_depth)
    {
    case 1:                          return 1;  // binary
    case 2: case 3: case 4:          return 4;  // 16
    case 5: case 6: case 7: case 8:  return 8;  // 256
    default: return 0;
    }
  }

  void get_color_table(uint8 *buf) const
  {
    for (uint i = 0; i < (1UL << m_color_depth); i++)
    {
      buf[i * 4    ] = m_clr_tbl[i * 3 + 2];
      buf[i * 4 + 1] = m_clr_tbl[i * 3 + 1];
      buf[i * 4 + 2] = m_clr_tbl[i * 3 + 0];
      buf[i * 4 + 3] = 0;
    }
  }

  int get_image(uint8 *out_buf, SPI_PROGRESS_FUNC prog_func, long pf_data);
};


// ----------------------------------------------------------------------------
//  vg^Cv錾
// ----------------------------------------------------------------------------

int check_image(const uint8 *data, size_t data_size);


#endif // gif_img_h__
