// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  local_mem.h
// ----------------------------------------------------------------------------

#ifndef local_mem_h__
#define local_mem_h__

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "conf.h"
#include <windows.h>


// ----------------------------------------------------------------------------
//  q[vǗNX
// ----------------------------------------------------------------------------

class local_memory
{
  bool    m_auto_free;
  bool    m_auto_unlock;

  HLOCAL  m_handle;
  void    *m_pointer;

public:
  local_memory(bool auto_free, bool auto_unlock)
       : m_auto_free(auto_free),
         m_auto_unlock(auto_unlock),
         m_handle(NULL),
         m_pointer(NULL) { }
  ~local_memory()
  {
    if (m_auto_unlock && m_pointer)  LocalUnlock(m_handle);
    if (m_auto_free && m_handle)     LocalFree(m_handle);
  }

  int lock(void)
  {
    if (NULL == m_pointer && m_handle)
      m_pointer = LocalLock(m_handle);
    return m_pointer ? 0 : 1;
  }

  void unlock(void)
  {
    if (m_pointer && m_handle)
    {
      LocalUnlock(m_handle);
      m_pointer = NULL;
    }
  }

  int alloc(size_t alloc_size)
  {
    free();
    m_handle = LocalAlloc(LMEM_MOVEABLE, alloc_size);
    return m_handle ? 0 : 1;
  }

  void free(void)
  {
    unlock();
    if (m_handle)
    {
      LocalFree(m_handle);
      m_handle = NULL;
    }
  }

  HLOCAL release(void)
  {
    unlock();
    HLOCAL  h = m_handle;
    m_handle = NULL;
    m_pointer = NULL;
    return h;
  }

  void *get_pointer(void) { return m_pointer; }
};

#endif // local_mem_h__
