// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  map_file.h
// ----------------------------------------------------------------------------

#ifndef map_file_h__
#define map_file_h__

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "conf.h"
#include "util.h"

#include <windows.h>


// ----------------------------------------------------------------------------
//  }bvt@C
// ----------------------------------------------------------------------------

class map_file
{
  HANDLE  m_fh;
  uint32  m_file_size;
  HANDLE  m_file_map;
  void    *m_map_addr;
  uint32  m_map_size;

  int create_file_map(HANDLE fh)
  {
    if (INVALID_HANDLE_VALUE == fh)
      return 1;

    m_file_size = GetFileSize(fh, NULL);
    HANDLE  file_map = CreateFileMapping(fh, NULL, PAGE_READONLY, 0, 0, NULL);
    if (NULL == file_map)
      return 1;

    m_file_map = file_map;

    return 0;
  }

  void unmap(void)
  {
    if (m_map_addr)
    {
      UnmapViewOfFile(m_map_addr);
      m_map_addr = NULL;
    }
  }

  void close_file_map(void)
  {
    if (m_file_map)
    {
      CloseHandle(m_file_map);
      m_file_map = NULL;
    }
  }

  void close_file(void)
  {
    if (INVALID_HANDLE_VALUE != m_fh)
    {
      CloseHandle(m_fh);
      m_fh = INVALID_HANDLE_VALUE;
    }
  }

public:
  map_file(void)
       : m_fh(INVALID_HANDLE_VALUE),
         m_file_map(NULL),
         m_map_addr(NULL) { }
  ~map_file() { close(); }

  int init(const char *filename)
  {
    close();
    m_fh = OpenReadFile(filename);
    return create_file_map(m_fh);
  }

  int init(HANDLE fh)
  {
    close();
    return create_file_map(fh);
  }

  void close(void)
  {
    unmap();
    close_file_map();
    close_file();
  }

  const void *map_all(void) { return map(m_file_size); }

  const void *map(uint map_size)
  {
    unmap();

    if (m_file_size < map_size)
      return NULL;

    if (NULL ==
        (m_map_addr = MapViewOfFile(m_file_map, FILE_MAP_READ, 0, 0, map_size)))
      return NULL;

    m_map_size = map_size;

    return m_map_addr;
  }

  uint32 get_mapped_size(void) const { return m_map_size; }
};


#endif // mapfile_h__
