// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  ifgif.spi - GIF import plug-in
//  (C) 2004 mij4x
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  spi_main.cpp
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "spi.h"
#include "in_data.h"
#include "map_file.h"
#include "local_mem.h"
#include "gif_img.h"


// ----------------------------------------------------------------------------
//  萔
// ----------------------------------------------------------------------------

#define PLUGIN_VERSION  "0.1"

const char *PLUGIN_INFO[] =
{
  "00IN",
  "GIF import plug-in ver." PLUGIN_VERSION " (C) 2004 mij4x",
  "*.gif",
  "GIF image",
};


// ----------------------------------------------------------------------------
//  GetPluginInfo()
// ----------------------------------------------------------------------------

int SPI_API GetPluginInfo(int info_no, LPSTR buf, int buf_len)
{
  int  copy_len = 0;

  if (0 <= info_no && (uint)info_no < countof(PLUGIN_INFO))
  {
    const char  *src = PLUGIN_INFO[info_no];
    char  *dst = buf;

    while (copy_len < buf_len)
    {
      copy_len++;
      if ('\0' == (*dst++ = *src++))
        break;
    }
  }

  return copy_len;
}


// ----------------------------------------------------------------------------
//  IsSupported()
// ----------------------------------------------------------------------------

int SPI_API IsSupported(LPSTR filename, DWORD data)
{
  uint8        buf[2000];
  const uint8  *chk_data;

  if (data & 0xFFFF0000)
    chk_data = (const uint8 *)data;
  else
  {
    DWORD  read_bytes;
    ReadFile((HANDLE)data, buf, 2000, &read_bytes, NULL);
    chk_data = buf;
  }

  return !check_image(chk_data, 2000);
}


// ----------------------------------------------------------------------------
//  GetPictureInfo()
// ----------------------------------------------------------------------------

int SPI_API GetPictureInfo(LPSTR buf, long len, unsigned int flag,
                           picture_info *pic_info)
{
  if (NULL == pic_info)
    return spi_object::SPI_ERROR_NOT_IMPLEMENTED;

  in_data  in((const uint8 *)buf, len, flag);
  if (NULL == in.get_pointer())
    return in.get_last_error_code();

  gif_image  img;
  if (img.open(in.get_pointer(), in.get_size()))
    return img.get_last_error_code();

  memset(pic_info, 0, sizeof(picture_info));

  pic_info->width = img.get_width();
  pic_info->height = img.get_height();
  pic_info->color_depth = img.get_color_depth();

  return spi_object::SPI_OK;
}


// ----------------------------------------------------------------------------
//  GetPicture()
// ----------------------------------------------------------------------------

int SPI_API GetPicture(LPSTR buf, long len, unsigned int flag,
                       HANDLE *bmi, HANDLE *img,
                       SPI_PROGRESS_FUNC prog_func, long data)
{
  if (NULL == bmi || NULL == img)
    return spi_object::SPI_ERROR_NOT_IMPLEMENTED;

  *bmi = *img = NULL;

  in_data  in((const uint8 *)buf, len, flag);
  if (NULL == in.get_pointer())
    return in.get_last_error_code();

  gif_image  img_file;
  if (img_file.open(in.get_pointer(), in.get_size()))
    return img_file.get_last_error_code();

  uint  img_size = CALC_DIB_ROW_BYTES(img_file.get_width(),
                                      img_file.get_color_depth())
                     * img_file.get_height();
  uint  pal_size = (img_file.get_color_depth() > 8)
                     ? 0 : (4 << img_file.get_color_depth());

  local_memory  bmi_mem(true, true), img_mem(true, true);
  if (img_mem.alloc(img_size) ||
      bmi_mem.alloc(sizeof(BITMAPINFOHEADER) + pal_size))
    return spi_object::SPI_ERROR_MEMORY_ALLOCATION;

  if (img_mem.lock() || bmi_mem.lock())
    return spi_object::SPI_ERROR_MEMORY;

  if (img_file.get_image((uint8 *)img_mem.get_pointer(), prog_func, data))
    return img_file.get_last_error_code();

  BITMAPINFOHEADER  *bi = (BITMAPINFOHEADER *)bmi_mem.get_pointer();
  memset(bi, 0, sizeof(BITMAPINFOHEADER));
  bi->biSize = sizeof(BITMAPINFOHEADER);
  bi->biWidth = img_file.get_width();
  bi->biHeight = img_file.get_height();
  bi->biBitCount = img_file.get_color_depth();
  bi->biPlanes = 1;
  img_file.get_color_table((uint8 *)(bi + 1));

  *bmi = bmi_mem.release();
  *img = img_mem.release();

  return spi_object::SPI_OK;
}


// ----------------------------------------------------------------------------
//  GetPreview()
// ----------------------------------------------------------------------------

int SPI_API GetPreview(LPSTR buf, long len, unsigned int flag,
                       HANDLE *bmi, HANDLE *img,
                       SPI_PROGRESS_FUNC prog_func, long data)
{
  return spi_object::SPI_ERROR_NOT_IMPLEMENTED;
}
